NAME = "hypertune"
VERSION = "1.2.1-rc1"
DESC = "A library for performing hyperparameter optimization with Polyaxon."
URL = "https://github.com/polyaxon/hypertune"
AUTHOR = "Polyaxon, Inc."
EMAIL = "contact@polyaxon.com"
LICENSE = "Apache 2.0"
